import os
from typing import List, Tuple
from urllib import request


class BibleDownloader:
    indexes = [
        "ACV.json.pbz2", "AKJV.json.pbz2", "AllEng.json.pbz2", "AllEs.json.pbz2", "AMP.json.pbz2",
        "ASV.json.pbz2", "BBE.json.pbz2", "BSB.json.pbz2", "BTX3.json.pbz2", "CSB.json.pbz2",
        "Darby.json.pbz2", "DRA.json.pbz2", "Dynamic.json.pbz2", "EBR.json.pbz2", "EsRV.json.pbz2",
        "ESV.json.pbz2", "ExtraEng.json.pbz2", "GNV.json.pbz2", "KJV%201611.json.pbz2",
        "KJV-like.json.pbz2", "KJV.json.pbz2", "Literal.json.pbz2", "Literal2.json.pbz2",
        "LSV.json.pbz2", "MSG.json.pbz2", "NASB%201995.json.pbz2", "NET.json.pbz2",
        "NIV%201984.json.pbz2", "NIV%202011.json.pbz2", "NIV.json.pbz2", "NKJV.json.pbz2",
        "NLT.json.pbz2", "RNKJV.json.pbz2", "RSV.json.pbz2", "RV1960.json.pbz2",
        "RV2004.json.pbz2", "RWV.json.pbz2", "UKJV.json.pbz2", "WEB.json.pbz2", "YLT.json.pbz2",
    ]
    __index_file_urls: List[Tuple[str, str]] = \
        [
            (f"https://github.com/samhaswon/bible_search/"
             f"raw/refs/heads/main/src/multi_bible_search/data/{x}",
             x
             )
            for x in indexes
        ]

    def download(self):
        base_dir = os.path.join(__file__[:-19], "data")
        os.makedirs(base_dir, exist_ok=True)
        for file_url, destination in self.__index_file_urls:
            destination = os.path.join(base_dir, destination.replace("%20", " "))
            print(f"Downloading {file_url} to {destination}")
            request.urlretrieve(file_url, destination)



"""
Test some of the performance aspects of the search module.
"""

from gc import get_referents
import os
import sys
import time
import unittest

from src.multi_bible_search.bible_search_adapter import BibleSearch


def getsize(obj):
    """
    Sum the size of an object and its members.
    """
    seen_ids = set()
    size = 0
    objects = [obj]
    while objects:
        need_referents = []
        for this_obj in objects:
            if id(this_obj) not in seen_ids:
                seen_ids.add(id(this_obj))
                size += sys.getsizeof(this_obj)
                need_referents.append(this_obj)
        objects = get_referents(*need_referents)
    return size


def get_directory_size(directory):
    """
    Get the size of all files in a directory.
    :param directory: The directory to search.
    :return: The size of the directory's files in bytes.
    """
    total_size = 0
    for dirpath, _, filenames in os.walk(directory):
        for filename in filenames:
            file_path = os.path.join(dirpath, filename)
            if os.path.isfile(file_path):
                total_size += os.path.getsize(file_path)
    return total_size


class TestPerf(unittest.TestCase):
    """
    Test searching performance.
    """
    def setUp(self) -> None:
        """
        Create the search object.
        """
        self.bible_search = BibleSearch()

    def test_perf(self):
        """
        Test the average case performance.
        """
        count = 200_000
        self.bible_search.load('KJV')
        # Warmup
        for _ in range(200):
            self.bible_search.search("Jesus wept")
        start = time.perf_counter()
        for _ in range(count):
            # Do some searching
            self.bible_search.search("Jesus wept")
        end = time.perf_counter()
        avg_time = (end - start) / count
        print(f"Total: {end - start:.4f}s\n"
              f"{avg_time:.8f}s per search average ({avg_time * 10 ** 6:.4f}μs)")
        start = time.perf_counter()
        for _ in range(count):
            # Do some searching
            self.bible_search.search("comest goest")
        end = time.perf_counter()
        avg_time = (end - start) / count
        print(f"Total: {end - start:.4f}s\n"
              f"{avg_time:.8f}s per search average ({avg_time * 10 ** 6:.4f}μs)")

    def test_perf_max_results(self):
        """
        Test the average case performance, but with limited results given to Python.
        """
        count = 200_000
        self.bible_search.load('KJV')
        start = time.perf_counter()
        for _ in range(count):
            # Do some searching
            self.bible_search.search("Jesus wept", max_results=100)
        end = time.perf_counter()
        avg_time = (end - start) / count
        print(f"Total: {end - start:.4f}s\n"
              f"{avg_time:.8f}s per search average ({avg_time * 10 ** 6:.4f}μs)")
        start = time.perf_counter()
        for _ in range(count):
            # Do some searching
            self.bible_search.search("comest goest", max_results=100)
        end = time.perf_counter()
        avg_time = (end - start) / count
        print(f"Total: {end - start:.4f}s\n"
              f"{avg_time:.8f}s per search average ({avg_time * 10 ** 6:.4f}μs)")

    def test_perf_long_query(self):
        """
        Test a couple of worst-case queries.
        """
        count = 50
        self.bible_search.load('KJV')
        start = time.perf_counter()
        for _ in range(count):
            # Do some searching
            self.bible_search.search(
                "Then were the king's scribes called at that time in the third month, that is, "
                "the month Sivan, on the three and twentieth day thereof; and it was written "
                "according to all that Mordecai commanded unto the Jews, and to the lieutenants, "
                "and the deputies and rulers of the provinces which are from India unto "
                "Ethiopia, an hundred twenty and seven provinces, unto every province according "
                "to the writing thereof, and unto every people after their language, and to the "
                "Jews according to their writing, and according to their language. Then were the "
                "king's scribes called at that time in the third month, that is, "
                "the month Sivan, on the three and twentieth day thereof; and it was written "
                "according to all that Mordecai commanded unto the Jews, and to the lieutenants, "
                "and the deputies and rulers of the provinces which are from India unto "
                "Ethiopia, an hundred twenty and seven provinces, unto every province according "
                "to the writing thereof, and unto every people after their language, and to the "
                "Jews according to their writing, and according to their language.",
                max_results=100)
        end = time.perf_counter()
        avg_time = (end - start) / count
        print(f"Excessively long query total: {end - start:.4f}s\n"
              f"{avg_time:.8f}s per search average")

        start = time.perf_counter()
        for _ in range(count):
            # Do some searching
            self.bible_search.search(
                "Then were the king's scribes called at that time in the third month, that is, "
                "the month Sivan, on the three and twentieth day thereof; and it was written "
                "according to all that Mordecai commanded unto the Jews, and to the lieutenants, "
                "and the deputies and rulers of the provinces which are from India unto "
                "Ethiopia, an hundred twenty and seven provinces, unto every province according "
                "to the writing thereof, and unto every people after their language, and to the "
                "Jews according to their writing, and according to their language.",
                max_results=100)
        end = time.perf_counter()
        avg_time = (end - start) / count
        print(f"Longest reasonable query total: {end - start:.4f}s\n"
              f"{avg_time:.8f}s per search average")

    def test_profile(self):
        """
        Test the memory usage of the search object with a single version and all versions loaded.
        """
        self.bible_search.load('KJV')
        new_kjv_size = \
            (getsize(self.bible_search) + self.bible_search.internal_index_size()) / (1024 ** 2)
        self.bible_search.load_all()
        new_all_size = \
            (self.bible_search.internal_index_size() + getsize(self.bible_search)) / (1024 ** 2)
        dir_size = get_directory_size("../src/multi_bible_search/data") / (1024 ** 2)
        print(f"KJV Only: {new_kjv_size:7.4f} MiB")
        print(f"All:      {new_all_size:7.4f} MiB")
        print(f"Disk:     {dir_size:7.4f} MiB")
        print(f"KJV reduction: {1106.2344970703125 - new_kjv_size:.4f} MiB")
        print(f"All reduction: {1106.2344970703125 - new_all_size:.4f} MiB")

    def test_average_keys(self):
        """
        Test the average retrieval speed of single keys.
        """
        count = 200
        with open("kjv_keys.txt", "r", encoding="utf-8") as key_file:
            keys = key_file.read().splitlines()
        self.bible_search.load('KJV')
        # warmup
        for _ in range(200):
            # Do some searching
            self.bible_search.search("Jesus wept")

        time_accumulator: float = 0.0

        for _ in range(count):
            for key in keys:
                start = time.perf_counter()
                result = self.bible_search.search(key)
                end = time.perf_counter()
                # Do something useful with the exhaustive search
                self.assertGreater(
                    len(result),
                    0,
                    msg=f"Couldn't find key {key} for some reason. "
                        f"Probably an issue in the C code."
                )
                time_accumulator += end - start
        avg_time = time_accumulator / count / len(keys)
        print(f"Total: {time_accumulator:.4f}s\n"
              f"{avg_time:.8f}s per search average ({avg_time * 10 ** 6:.4f}μs)")


if __name__ == '__main__':
    unittest.main()



"""
Exception raised when an invalid version is passed to the search object.
"""


class InvalidVersion(Exception):
    """
    Raised when an invalid version is passed to the search object.
    """
    def __init__(self, version_attempt: str):
        super().__init__(f"Invalid version {version_attempt}")



"""
Adapter class for the C implementation of the search engine.
This mostly just makes it easier to work with the module
by automatically managing version indexes which the C class does not do.
"""
import bz2
import os
import sys
from typing import List, Union

# PyCharm and Pylint both can't figure this one out,
# but it works and is correct.
# pylint: disable=no-name-in-module
from .multi_bible_search import BibleSearch as cBibleSearch
from .invalid_version import InvalidVersion


class BibleSearch:
    """
    Search versions of the Bible
    """
    def __init__(self, preload: Union[List[str], None] = None):
        """
        :param preload: List of versions to preload.
        """
        # (C) Search object
        self.__c_search = cBibleSearch()

        try:
            if len(os.listdir(os.path.join(__file__[:-23], "data"))) < 40:
                from .bible_downloader import BibleDownloader

                downloader = BibleDownloader()
                downloader.download()
        except FileNotFoundError:
            from .bible_downloader import BibleDownloader

            downloader = BibleDownloader()
            downloader.download()

        # Common sets
        self.__dynamic: set = {"CSB", "NLT", "NET"}
        self.__kjv_like: set = {"AKJV", "GNV", "KJV", "KJV 1611", "RNKJV", "UKJV"}
        self.__literal: set = {"ACV", "AMP", "ASV", "ESV", "NASB 1995", "NKJV", "RSV", "RWV", "WEB"}
        self.__literal2: set = {"BSB", "LSV", "YLT"}
        self.__niv: set = {"NIV 1984", "NIV 2011"}
        self.__esrv: set = {'RV1960', 'RV2004'}
        self.__extra_english: set = {"Darby", "EBR"}

        # Language sets
        self.__english_versions: set = {'ACV', 'AKJV', 'AMP', 'ASV', 'BBE', 'BSB', 'CSB', 'Darby',
                                        'DRA', 'EBR', 'ESV', 'GNV', 'KJV', 'KJV 1611', 'LSV',
                                        'MSG', 'NASB 1995', 'NET', 'NIV 1984', 'NIV 2011', 'NKJV',
                                        'NLT', 'RNKJV', 'RSV', 'RWV', 'UKJV', 'WEB', 'YLT'}
        self.__spanish_versions: set = {'BTX3', 'RV1960', 'RV2004'}

        # Every supported version
        self.__versions: set = self.__english_versions | self.__spanish_versions

        # Preload common index
        self._load_version("AllEng", preload=True)

        # What is currently stored in C
        self.__loaded: set = set()
        self.__preloaded: set = set()

        if preload:
            for version in preload:
                self.load(version)

    def _load_version(self, version: str, preload: bool = False) -> None:
        """
        Preloads a given version's search index.
        It assumes that the version is valid.
        :param version: The version to preload.
        :return: None
        """
        base_path = os.path.dirname(os.path.abspath(__file__))
        with bz2.open(f"{base_path}/data/{version}.json.pbz2", "rt", encoding='utf-8') as data_file:
            self.__c_search.load(data_file.read(), version)
        if not preload:
            self.__loaded.add(version)

    def load(self, version: str) -> None:
        """
        Preloads a given version's search index.
        :param version: The version to preload.
        :return: None
        :raises KeyError: For invalid version strings.
        """
        # Quick check that the version is valid
        if version not in self.__versions:
            raise InvalidVersion(version)

        # load Spanish common index if applicable
        if version in self.__spanish_versions and "AllEs" not in self.__preloaded:
            self._load_version("AllEs", preload=True)
            self.__preloaded.add("AllEs")

        # Load KJV-like common index if applicable
        if version in self.__kjv_like and "KJV-like" not in self.__preloaded:
            self._load_version("KJV-like", preload=True)
            self.__preloaded.add("KJV-like")

        # Load NIV common index if applicable
        elif version in self.__niv and "NIV" not in self.__preloaded:
            self._load_version("NIV", preload=True)
            self.__preloaded.add("NIV")

        # Load Literal common index if applicable
        elif version in self.__literal and "Literal" not in self.__preloaded:
            self._load_version("Literal", preload=True)
            self.__preloaded.add("Literal")

        # Load Literal2 common index if applicable
        elif version in self.__literal2 and "Literal2" not in self.__preloaded:
            self._load_version("Literal2", preload=True)
            self.__preloaded.add("Literal2")

        # Load Dynamic common index if applicable
        elif version in self.__dynamic and "Dynamic" not in self.__preloaded:
            self._load_version("Dynamic", preload=True)
            self.__preloaded.add("Dynamic")

        # Load the common index of some common Spanish versions
        elif version in self.__esrv and "EsRV" not in self.__preloaded:
            self._load_version("EsRV", preload=True)
            self.__preloaded.add("EsRV")

        # Load the index of some assorted English versions
        elif version in self.__extra_english and "ExtraEng" not in self.__preloaded:
            self._load_version("ExtraEng", preload=True)
            self.__preloaded.add("ExtraEng")

        # Load the version
        self._load_version(version)

    def load_all(self) -> None:
        """
        Preload all version indices.
        """
        for version in self.__versions:
            self.load(version)

    def unload_version(self, version: str) -> None:
        """
        Unload a version's index from memory.
        :param version: The version to remove.
        :return: None
        :raises InvalidVersion: If the version is invalid or not loaded, raises an exception.
        """
        if version in self.__versions and version in self.__loaded:
            self.__c_search.unload(version)
            self.__loaded.remove(version)
        else:
            raise InvalidVersion(version)

    def search(self, query: str, version: str = "KJV", max_results: int = sys.maxsize) -> List[str]:
        """
        Search for a passage in the Bible.
        :param query: The search query string.
        :param version: The version to search.
        :param max_results: The maximum number of results to retrieve.
        :return: List of match references (e.g., `["John 11:35", "Matthew 1:7", ...]`).
        """
        # Load the version if it is not already loaded
        if version not in self.__loaded:
            self.load(version)
        return self.__c_search.search(query, version, max_results)

    def internal_index_size(self) -> int:
        """
        Gets the size of the index stored in C in bytes.
        :return: Index size in bytes.
        """
        return self.__c_search.index_size()

    @property
    def loaded(self) -> List[str]:
        """
        A list of the versions currently loaded in the search object.
        This does not include common indices.
        """
        return list(self.__loaded)

    @property
    def versions(self) -> List[str]:
        """
        A list of versions available in the search index.
        """
        return list(self.__versions)

    @property
    def english_versions(self) -> List[str]:
        """
        A list of English versions available in the search index.
        """
        return list(self.__english_versions)

    @property
    def spanish_versions(self) -> List[str]:
        """
        A list of Spanish versions available in the search index.
        """
        return list(self.__spanish_versions)



"""
Translation layer
"""

books = {
        'Genesis': 1, 'Exodus': 2, 'Leviticus': 3, 'Numbers': 4, 'Deuteronomy': 5, 'Joshua': 6,
        'Judges': 7, 'Ruth': 8, '1 Samuel': 9, '2 Samuel': 10, '1 Kings': 11, '2 Kings': 12,
        '1 Chronicles': 13, '2 Chronicles': 14, 'Ezra': 15, 'Nehemiah': 16, 'Esther': 17,
        'Job': 18, 'Psalms': 19, 'Proverbs': 20, 'Ecclesiastes': 21, 'Song of Solomon': 22,
        'Isaiah': 23, 'Jeremiah': 24, 'Lamentations': 25, 'Ezekiel': 26, 'Daniel': 27, 'Hosea': 28,
        'Joel': 29, 'Amos': 30, 'Obadiah': 31, 'Jonah': 32, 'Micah': 33, 'Nahum': 34,
        'Habakkuk': 35, 'Zephaniah': 36, 'Haggai': 37, 'Zechariah': 38, 'Malachi': 39,
        'Matthew': 40, 'Mark': 41, 'Luke': 42, 'John': 43, 'Acts': 44, 'Romans': 45,
        '1 Corinthians': 46, '2 Corinthians': 47, 'Galatians': 48, 'Ephesians': 49,
        'Philippians': 50, 'Colossians': 51, '1 Thessalonians': 52, '2 Thessalonians': 53,
        '1 Timothy': 54, '2 Timothy': 55, 'Titus': 56, 'Philemon': 57, 'Hebrews': 58, 'James': 59,
        '1 Peter': 60, '2 Peter': 61, '1 John': 62, '2 John': 63, '3 John': 64, 'Jude': 65,
        'Revelation': 66
    }

rbooks = {
        1: 'Genesis', 2: 'Exodus', 3: 'Leviticus', 4: 'Numbers', 5: 'Deuteronomy', 6: 'Joshua',
        7: 'Judges', 8: 'Ruth', 9: '1 Samuel', 10: '2 Samuel', 11: '1 Kings', 12: '2 Kings',
        13: '1 Chronicles', 14: '2 Chronicles', 15: 'Ezra', 16: 'Nehemiah', 17: 'Esther',
        18: 'Job', 19: 'Psalms', 20: 'Proverbs', 21: 'Ecclesiastes', 22: 'Song of Solomon',
        23: 'Isaiah', 24: 'Jeremiah', 25: 'Lamentations', 26: 'Ezekiel', 27: 'Daniel', 28: 'Hosea',
        29: 'Joel', 30: 'Amos', 31: 'Obadiah', 32: 'Jonah', 33: 'Micah', 34: 'Nahum',
        35: 'Habakkuk', 36: 'Zephaniah', 37: 'Haggai', 38: 'Zechariah', 39: 'Malachi',
        40: 'Matthew', 41: 'Mark', 42: 'Luke', 43: 'John', 44: 'Acts', 45: 'Romans',
        46: '1 Corinthians', 47: '2 Corinthians', 48: 'Galatians', 49: 'Ephesians',
        50: 'Philippians', 51: 'Colossians', 52: '1 Thessalonians', 53: '2 Thessalonians',
        54: '1 Timothy', 55: '2 Timothy', 56: 'Titus', 57: 'Philemon', 58: 'Hebrews', 59: 'James',
        60: '1 Peter', 61: '2 Peter', 62: '1 John', 63: '2 John', 64: '3 John', 65: 'Jude',
        66: 'Revelation'
    }


def translate(book: str, chapter: int, verse: int) -> int:
    """
    Get the (from 1) index of a book of the Bible.
    :param book: The book name.
    :param chapter: The chapter reference.
    :param verse: The verse reference.
    :return: Index of the book.
    """
    # quick bounds check
    assert verse <= 176
    assert chapter <= 150
    return books[book] * 1_000_000 + chapter * 1_000 + verse


def rtranslate(reference: int) -> str:
    """
    Gets the name of a book based on its index. Opposite of translate()
    :param reference: The reference to translate.
    :return: The reference as a string.
    """
    return f"{rbooks[reference // 1_000_000]} {reference // 1_000 % 1_000}:{reference % 1000}"



"""
Set up the multi-bible-search extension.
"""
import os

# pylint: disable=import-error
from setuptools import setup, Extension


if os.name == "nt":
    # MSVC flags
    flags = ["/favor:blend",  # Optimize for AMD and Intel
             "/GF",  # (Eliminate Duplicate Strings)
             "/Qpar",  # (Auto-parallelizer)
             "/arch:AVX2",  # For some reason, AVX2 is faster than AVX512 in my testing.
             "/fp:fast"  # Faster floating point ops, since precision is not needed for this
             ]
else:
    flags = ["-O3", "-mavx2"]

setup(
    name='multi_bible_search',
    ext_modules=[
        Extension('multi_bible_search.multi_bible_search',
                  ['src/multi_bible_search/multi_bible_search.c'],
                  include_dirs=['src/multi_bible_search/'],
                  extra_compile_args=flags,
                  )
    ],
)



"""
Test the functionality of the search module.
"""
import unittest

# pylint: disable=import-error
from src.multi_bible_search.bible_search_adapter import BibleSearch
from src.multi_bible_search.invalid_version import InvalidVersion


class TestSearch(unittest.TestCase):
    """
    Test the functionality of the search module.
    """
    def setUp(self) -> None:
        """
        Make the Bible search object.
        :return: None.
        """
        self.bible_search = BibleSearch()

    def test_query_a(self):
        """
        Test some basic queries.
        :return: None.
        """
        query1 = self.bible_search.search("Jesus wept")
        self.assertIn("John 11:35", query1)
        self.assertIn("John 11:35", query1[:15])
        query2 = self.bible_search.search(
            "hospitality, a lover of good men, sober, just, holy, temperate"
        )
        self.assertIn("Titus 1:8", query2)
        query3 = self.bible_search.search("elect")
        self.assertIn("Titus 1:1", query3)
        self.assertIn("Romans 8:33", query3)
        query4 = self.bible_search.search("Caesarea Philippi")
        self.assertIn("Mark 8:27", query4)
        self.assertIn("Mark 8:27", query4[:5])
        query5 = self.bible_search.search("scorneth")
        self.assertIn("Proverbs 19:28", query5)

    def test_query_b(self):
        """
        Test some more complicated queries.
        :return: None.
        """
        query6 = self.bible_search.search("Therefore came I forth to meet thee")
        self.assertIn("Proverbs 7:15", query6)
        # Sorting is backwards for partial matches in the C version
        self.assertIn("Proverbs 7:15", query6[:5])
        query7 = self.bible_search.search(
            "And Israel dwelt in the land of Egypt, in the country of Goshen; and they "
            "had possessions therein, and grew, and multiplied exceedingly.")
        self.assertIn("Genesis 47:27", query7[:3])
        query8 = self.bible_search.search("notawordinthebible")
        self.assertEqual(len(query8), 0)
        query9 = self.bible_search.search("Jesus wept", "ESV")
        self.assertGreater(len(query9), 0)
        query10 = self.bible_search.search(
            "And your feet shod with the preparation of the gospel of peace"
        )
        self.assertEqual("Ephesians 6:15", query10[0])

    def test_query_edge_cases(self):
        """
        Test query edge cases.
        :return: None.
        """
        # Test it with some unicode
        query11 = self.bible_search.search("something æ\u200B\u201Dב” else")
        self.assertGreater(len(query11), 0)
        query12 = self.bible_search.search("wh")
        self.assertEqual(len(query12), 0)
        query13 = self.bible_search.search("w")
        self.assertEqual(len(query13), 0)
        query14 = self.bible_search.search("a")
        self.assertGreater(len(query14), 0)
        query15 = self.bible_search.search("")
        self.assertEqual(len(query15), 0)

    def test_versions(self):
        """
        Make sure the versions are placed correctly.
        :return: None.
        """
        # Basic version getter
        self.assertIn("KJV", self.bible_search.versions)
        self.assertIn("ESV", self.bible_search.versions)
        self.assertIn("ASV", self.bible_search.versions)

        # Language version getter
        self.assertIn("KJV", self.bible_search.english_versions)
        self.assertIn("ESV", self.bible_search.english_versions)
        self.assertIn("ASV", self.bible_search.english_versions)

        self.assertIn("BTX3", self.bible_search.spanish_versions)
        self.assertIn("RV1960", self.bible_search.spanish_versions)
        self.assertIn("RV2004", self.bible_search.spanish_versions)

        self.assertNotIn("KJV", self.bible_search.spanish_versions)
        self.assertNotIn("BTX3", self.bible_search.english_versions)
        self.assertNotIn("RV1960", self.bible_search.english_versions)
        self.assertNotIn("RV2004", self.bible_search.english_versions)

    def test_load_all(self):
        """
        Test loading all versions.
        """
        self.bible_search.load_all()
        self.assertEqual(self.bible_search.loaded.sort(), self.bible_search.versions.sort())
        english_versions = self.bible_search.english_versions
        english_versions.sort()
        for version in english_versions:
            # print(version)
            result = self.bible_search.search("jesus wept", version)
            self.assertGreater(len(result), 0, f"Fail for version {version}")

        spanish_versions = self.bible_search.spanish_versions
        spanish_versions.sort()
        for version in spanish_versions:
            result = self.bible_search.search("Jesús lloró", version)
            self.assertGreater(
                len(result),
                0,
                f"Fail for version {version}.\n"
                f"Loaded: {self.bible_search.loaded}\n"
            )

    def test_unload_version(self):
        """
        Make sure that unloading a version index works correctly.
        :return:
        """
        self.bible_search.load("KJV")
        self.bible_search.unload_version("KJV")

        # Should raise an Exception if we try to unload it again.
        with self.assertRaises(InvalidVersion):
            self.bible_search.unload_version("KJV")
        self.bible_search.load("KJV")
        # This should be skipped and not cause an error
        self.bible_search.load("KJV")


if __name__ == '__main__':
    unittest.main()
